// HACK! Ignore this if used as a top-level compile target.
#ifdef OPEN_LISTS_DIVERSE_OPEN_LIST_H
#include "../rng.h"
#include "../state_registry.h"

#include <algorithm>
#include <functional>
#include <limits>
#include <cmath>

using namespace std;

template<class Entry>
DiverseOpenList<Entry>::DiverseOpenList(const Options &opts)
    : OpenList<Entry>(opts.get<bool>("pref_only")),
      global_open_list(opts.get<OpenList<Entry> *>("open")),
      local_open_list(opts.get<OpenList<Entry> *>("local_open")),
      num_remaining_steps(1) { // set to one to make --num_remaining_steps == 0 in remove min
    set<Heuristic *> local_hset;
    local_open_list->get_involved_heuristics(local_hset);
    local_heuristics.insert(local_heuristics.begin(), local_hset.begin(), local_hset.end());

    set<Heuristic *> global_hset;
    global_open_list->get_involved_heuristics(global_hset);
    global_heuristics.insert(global_heuristics.begin(), global_hset.begin(), global_hset.end());
}

template<class Entry>
DiverseOpenList<Entry>::~DiverseOpenList() {
}

template<class Entry>
StateID DiverseOpenList<Entry>::get_stateid_from_entry(const Entry &) {
    ABORT("Not implemented for Entry");
    return StateID::no_state;
}

template<class Entry>
void DiverseOpenList<Entry>::evaluate(int g, bool preferred) {
    last_g_value = g;
    local_open_list->evaluate(g, preferred);
}

template<class Entry>
bool DiverseOpenList<Entry>::is_dead_end() const {
    return local_open_list->is_dead_end();
}

template<class Entry>
bool DiverseOpenList<Entry>::dead_end_is_reliable() const {
    return local_open_list->dead_end_is_reliable();
}

template<class Entry>
void DiverseOpenList<Entry>::merge_open_lists() {
    while (!local_open_list->empty()) {
        Entry entry = local_open_list->remove_min();
        StateID id = get_stateid_from_entry(entry);
        GlobalState state = g_state_registry->lookup_state(id);
        for (vector<Heuristic *>::iterator it = global_heuristics.begin(); it != global_heuristics.end(); ++it) {
            (*it)->evaluate(state);
        }
        global_open_list->evaluate(state_id_to_g[id], false);
        global_open_list->insert(entry);
    }
}

template<class Entry>
void DiverseOpenList<Entry>::reset_local_open_list() {
    merge_open_lists();
    Entry entry = global_open_list->remove_min();
    StateID id = get_stateid_from_entry(entry);
    GlobalState state = g_state_registry->lookup_state(id);
    for (vector<Heuristic *>::iterator it = local_heuristics.begin(); it != local_heuristics.end(); ++it) {
        (*it)->evaluate(state);
    }
    local_open_list->evaluate(state_id_to_g[id], false);
    local_open_list->insert(entry);
    num_remaining_steps = local_heuristics[0]->get_heuristic();
}

template<class Entry>
void DiverseOpenList<Entry>::get_involved_heuristics(set<Heuristic *> &hset) {
    local_open_list->get_involved_heuristics(hset);
}

template<class Entry>
int DiverseOpenList<Entry>::insert(const Entry &entry) {
    int count = local_open_list->insert(entry);
    if (count > 0) {
        state_id_to_g[get_stateid_from_entry(entry)] = last_g_value;
    }
    return count;
}

template<class Entry>
Entry DiverseOpenList<Entry>::remove_min(vector<int> *key) {
    assert(!empty());
    if (--num_remaining_steps == 0 || local_open_list->empty()) {
        reset_local_open_list();
    }
    Entry result = local_open_list->remove_min(key);
    state_id_to_g.erase(get_stateid_from_entry(result));
    return result;
}

template<class Entry>
bool DiverseOpenList<Entry>::empty() const {
    return global_open_list->empty() && local_open_list->empty();
}

template<class Entry>
void DiverseOpenList<Entry>::clear() {
    global_open_list->clear();
    local_open_list->clear();
}

// TODO: we have to inline because this file is included multiple times
template<>
inline StateID DiverseOpenList<StateID>::get_stateid_from_entry(const StateID &id) {
    return id;
}

//#include "lazy_search.h"
typedef pair<StateID, const GlobalOperator *> OpenListEntryLazy;
template<>
inline StateID DiverseOpenList<OpenListEntryLazy>::get_stateid_from_entry(const OpenListEntryLazy &id) {
    return id.first;
}

template<class Entry>
OpenList<Entry> *DiverseOpenList<Entry>::_parse(OptionParser &parser) {
    parser.document_synopsis("Diverse best first search",
                             "Diverse best first search diversifies search by "
                             "probabalistic selecting nodes which not have the best heuristic value");
    parser.add_option<OpenList<Entry> *>("open", "Open list used for all open nodes");
    parser.add_option<OpenList<Entry> *>("local_open", "Open list used for the local search");

    parser.add_option<bool>(
        "pref_only",
        "insert only nodes generated by preferred operators", "false");

    Options opts = parser.parse();

    DiverseOpenList<Entry> *engine = 0;
    if (!parser.dry_run()) {
        engine = new DiverseOpenList<Entry>(opts);
    }

    return engine;
}

#endif
