#ifndef OPEN_LISTS_RANDOM_BUCKETS_OPEN_LIST_H
#define OPEN_LISTS_RANDOM_BUCKETS_OPEN_LIST_H

#include "standard_scalar_open_list.h"

/*
  Random open list that uses a single evaluator.
  With probability epsilon the next entry is selected with a uniform random distribution, otherwise the minimum entry is chosen.
  This implementation is similar to the original implementation by Valenzano et al. (see below) in that it extends StandardScalarOpenList (Valenzano, personal communications).
  While it is fast to remove the minimum entry (O(1)) the removal of a random entry is expensive (O(n) in the number of buckets).
  Therefore, we also provide a second implementation with better worst-case guarantees in random_open_list.h.
  Based on:
    Richard Valenzano, Nathan R. Sturtevant, Jonathan Schaeffer, and Fan Xie.
    A comparison of knowledge-based GBFS enhancements and knowledge-free exploration.
    In Proceedings of the Twenty-Fourth International Conference on Automated
    Planning and Scheduling (ICAPS 2014), pages 375-379. AAAI Press, 2014.
*/

template<class Entry>
class RandomBucketsOpenList : public StandardScalarOpenList<Entry>{
    typedef std::deque<Entry> Bucket;
    double epsilon;

public:
    explicit RandomBucketsOpenList(const Options &opts);
    ~RandomBucketsOpenList();

    virtual Entry remove_min(std::vector<int> *key = 0);

    static OpenList<Entry> *_parse(OptionParser &parser);
};

#include "random_buckets_open_list.cc"

// HACK! Need a better strategy of dealing with templates, also in the Makefile.
#endif
