// HACK! Ignore this if used as a top-level compile target.
#ifdef OPEN_LISTS_RANDOM_OPEN_LIST_H

#include "../globals.h"
#include "../rng.h"
#include "../utilities.h"

#include <algorithm>

using namespace std;


template<class Entry>
bool RandomOpenList<Entry>::GreaterThanComparator::operator()(
    const pair<RandomHeapKey, Entry> &a,
    const pair<RandomHeapKey, Entry> &b) const {
const RandomHeapKey &key_a = a.first;
const RandomHeapKey &key_b = b.first;
return (key_a.h > key_b.h) || (key_a.h == key_b.h && key_a.id > key_b.id);
}

template<typename Heap, typename Comparator>
static void adjust_heap_up(Heap &heap, size_t pos, const Comparator &comp) {
    assert(in_bounds(pos, heap));
    while (pos != 0) {
        size_t parent_pos = (pos - 1) / 2;
        if (!comp(heap[parent_pos], heap[pos]))
            break;
        swap(heap[pos], heap[parent_pos]);
        pos = parent_pos;
    }
}

template<class Entry>
RandomOpenList<Entry>::RandomOpenList(const Options &opts)
    : OpenList<Entry>(opts.get<bool>("pref_only")),
      evaluator(opts.get<ScalarEvaluator *>("eval")),
      epsilon(opts.get<double>("epsilon")),
      size(0),
      next_id(0) {
}

template<class Entry>
RandomOpenList<Entry>::~RandomOpenList() {
}

template<class Entry>
int RandomOpenList<Entry>::insert(const Entry &entry) {
    if (OpenList<Entry>::only_preferred && !last_preferred)
        return 0;
    if (dead_end)
        return 0;
    RandomHeapKey key = {
        next_id++, last_evaluated_value
    };
    open_list.push_back(make_pair(key, entry));
    push_heap(open_list.begin(), open_list.end(), comparator);
    ++size;
    return 1;
}

template<class Entry>
Entry RandomOpenList<Entry>::remove_min(vector<int> *key) {
    assert(size > 0);
    if (g_rng() < epsilon) {
        int pos = g_rng(size);
        open_list[pos].first.h = -numeric_limits<int>::max();
        adjust_heap_up(open_list, pos, comparator);
    }
    pop_heap(open_list.begin(), open_list.end(), comparator);
    pair<RandomHeapKey, Entry> key_and_entry = open_list.back();
    open_list.pop_back();
    if (key) {
        assert(key->empty());
        key->push_back(key_and_entry.first.h);
    }
    --size;
    return key_and_entry.second;
}

template<class Entry>
bool RandomOpenList<Entry>::empty() const {
    return size == 0;
}

template<class Entry>
void RandomOpenList<Entry>::clear() {
    open_list.clear();
    size = 0;
}

template<class Entry>
bool RandomOpenList<Entry>::is_dead_end() const {
    return dead_end;
}

template<class Entry>
bool RandomOpenList<Entry>::dead_end_is_reliable() const {
    return dead_end_reliable;
}

template<class Entry>
void RandomOpenList<Entry>::evaluate(int g, bool preferred) {
    get_evaluator()->evaluate(g, preferred);
    last_evaluated_value = get_evaluator()->get_value();
    last_preferred = preferred;
    dead_end = get_evaluator()->is_dead_end();
    dead_end_reliable = get_evaluator()->dead_end_is_reliable();
}

template<class Entry>
void RandomOpenList<Entry>::get_involved_heuristics(set<Heuristic *> &hset) {
    evaluator->get_involved_heuristics(hset);
}

template<class Entry>
OpenList<Entry> *RandomOpenList<Entry>::_parse(OptionParser &parser) {
    parser.document_synopsis(
        "Random open list",
        "With probability epsilon this open list chooses the next entry uniformly "
        "randomly, otherwise it returns the minimum entry.");
    parser.add_option<ScalarEvaluator *>("eval", "scalar evaluator");
    parser.add_option<bool>(
        "pref_only",
        "insert only nodes generated by preferred operators", "false");
    parser.add_option<double>(
        "epsilon",
        "probability between 0 and 1 with which to choose the next entry randomly", "0.05");
    Options opts = parser.parse();

    if (parser.dry_run()) {
        return 0;
    } else {
        double epsilon = opts.get<double>("epsilon");
        if (epsilon > 1 || epsilon < 0) {
            parser.error("epsilon must be between 0 and 1");
        }
        return new RandomOpenList<Entry>(opts);
    }
}
#endif
